package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_RIGHTS_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_TYPE_SHARED;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_COMMUNITY;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_PROJECT;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.UNSAVED_ID;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.AccessControlException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.Calendar;
import com.fmi.weprom.api.Community;
import com.fmi.weprom.api.File;
import com.fmi.weprom.api.Permission;
import com.fmi.weprom.api.Role;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.WEPROMException;
import com.fmi.weprom.util.WepromRequestCycleUtil;
import static com.fmi.weprom.api.Project.*;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.*;

public class CommunityImpl extends WorkspaceImpl implements Community {
  
  private static final String CACHE_NAME = "CommunityCache";
  private static final Log LOG = LogFactory.getLog(CommunityImpl.class);
  private static final int INVALID_USER = -1;
  
  protected int owner_id = INVALID_USER;
  protected String name;
  protected transient CalendarImpl calendar;
  protected transient WorkspaceInfo info;
  private Integer _ID = null;
  
  
  private CommunityImpl(User owner, String name, int id) {
    this(owner.getID(), name, id);
  }
  
  private CommunityImpl(int owner_id, String name, int id) {
    super();
    this.owner_id = owner_id;
    this.name = name;
    this.id = id;  
  }
  
  CommunityImpl(User user) {
    this(user, null, UNSAVED_ID);
  }
  
  public Calendar getCalendar() {
    if (calendar==null) {
      calendar = new CalendarImpl(this);
    }
    return calendar;
  }

  public void save() throws WEPROMException {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Community already exists");
    }
    if (name==null) {
      throw new IllegalStateException("Community does not have name");
    }
    LOG.debug("save() "+"id="+id + " user.id="+owner_id);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    boolean transactionOK = false;
    try {
      conn = Util.currentSession().getConnection();
      LOG.debug("save() "+"BEGIN TRANSACTION: CREATE_COMMUNITY");
      // Disable auto commit
      conn.setAutoCommit(false);
      
      // check if the community name is not alreay in use
      if (isCommunityNameFree(conn, name)) {  
        pstmt = conn.prepareStatement("INSERT INTO " + MASTER_TABLE + " (type,owner_id,access_type,name,creation_date) VALUES(?,?,?,?,NOW())",
                                      PreparedStatement.RETURN_GENERATED_KEYS);
        int i = 0;
        pstmt.setInt(++i, MASTER_TYPE_COMMUNITY);
        pstmt.setInt(++i, owner_id);
        pstmt.setInt(++i, ACCESS_TYPE_SHARED);
        pstmt.setString(++i, name);
        
        pstmt.executeUpdate();
        
        WorkspaceInfo info = getInfo();
        
        // get autogenerated project id
        rs = pstmt.getGeneratedKeys();
        if (rs.next()) {
          int communityID = rs.getInt(1);
          this.id = communityID; 
          LOG.debug("save() generated id="+id);
          
          grantOwnerRoles();
          
          // insert into workspaces
          //WorkspaceInfo info = createInfo();          
          info.save();
          
          // Commit updates
          conn.commit();
          transactionOK = true;
          LOG.debug("save() END TRANSACTION: CREATE_COMMUNITY");          
          cache();          
        }
        
      } else {
        throw new WEPROMException("Faild to create community - name is already in use");
      }
      
    } catch (SQLException e) {
      LOG.error("save() "+"FAILED", e);
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("save() ROLLBACK TRANSACTION: CREATE_COMMUNITY");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}        
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
  }

  public void update() throws WEPROMException {
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Community does not exists");
    }
    User caller = WepromRequestCycleUtil.currentUser();
    try {
      getInfo().update();
    } catch (SQLException e) {
      throw new WEPROMException("FAILED to update community: "+toString(), e);
    }
  }

  public void delete() throws WEPROMException {
    // TODO Auto-generated method stub
   
    uncache();
  }
  
  public String getName() {
    return name;
  }

  public void setName(String name) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Community already exists - name could not be changed");
    }
    this.name = name;   
  }

  public Date getStartDate() {
    return getInfo().getStartDate();
  }

  public Date getEndDate() {
    return getInfo().getEndDate();
  }

  public String getDescription() {
    return getInfo().getDescription();
  }

  public int getCompleted() {
    return getInfo().getCompleted();
  }

  public int getStatus() {
    return getInfo().getStatus();
  }

  public void setStartDate(Date startDate) {
    getInfo().setStartDate(startDate);
  }

  public void setEndDate(Date endDate) {
    getInfo().setEndDate(endDate);
  }

  public void setDescription(String description) {
    getInfo().setDescription(description);
  }

  public void setCompleted(int completed) {
    getInfo().setCompleted(completed);
  }

  public void setStatus(int status) {
    getInfo().setStatus(status);
  }

  public User getOwner() {
    if (owner_id!=INVALID_USER) {
      return UserImpl.getInstance(owner_id);
    }
    return null;
  }

  public List getUsers() {
    return WorkspaceImpl.getUsers(LOG, id);
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////
  
  int getType() {
    return MASTER_TYPE_COMMUNITY;
  }
  
  private WorkspaceInfo createInfo() {
    return new WorkspaceInfo(this);
  }
  
  private WorkspaceInfo getInfo() {
    if (info==null) {
      if (id==UNSAVED_ID) {
        info = createInfo();
      } else {
        info = WorkspaceInfo.load(this);
      }
    }
    return info;
  }
  
  private Integer get_ID() {
    if (id==UNSAVED_ID) return null;
    if (_ID==null) {
      _ID = new Integer(id);
    }
    return _ID;
  }
  
  private void uncache() {
    LOG.debug("uncache: " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.remove(get_ID());
  }
  
  private void cache() {
    LOG.debug("cache: " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.put(new Integer(id), this);
  }
  
  static CommunityImpl getCached(int id) {
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    Integer _id = new Integer(id);
    CommunityImpl community = (CommunityImpl) cache.get(_id);
    return community;
  }
  
  private static CommunityImpl load(int id) {
    String _method = "load(" + id + ") "; 
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT name, owner_id FROM "+MASTER_TABLE+ " WHERE id=?"); 
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      while (rs.next()) {
        int i = 0;
        String name = rs.getString(++i);
        int owner_id = rs.getInt(++i);
        CommunityImpl p = CommunityImpl.getInstance(owner_id, name, id);
        return p;
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;    
  }
  
  static CommunityImpl getInstance(int owner_id, String name, int id) {
    CommunityImpl community = getCached(id);
    if (community==null) {
      community = new CommunityImpl(owner_id, name, id);
      LOG.debug("getInstance() "+"create new : " + community);
      community.cache();
    } else {
      LOG.debug("getInstance() "+"load cached : " + community);
    }
    return community;    
  }
  
  static CommunityImpl getInstance(int id) {
    CommunityImpl community = getCached(id);
    if (community!=null) {
      return community;
    }
    return load(id);
  }
  
  static List loadCommunities() {
    LOG.debug("loadCommunities()");
    List result = new ArrayList();
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT id, name, owner_id FROM "+MASTER_TABLE+ " where type=?");
      pstmt.setInt(1, MASTER_TYPE_COMMUNITY);
      rs = pstmt.executeQuery();
      while (rs.next()) {
        int id = rs.getInt(1);
        String name = rs.getString(2);
        int owner_id = rs.getInt(3);
        //User owner = UserImpl.getUser(weprom, owner_id);
        CommunityImpl c = CommunityImpl.getInstance(owner_id, name, id);
        result.add(c);
      }
    } catch (SQLException e) {
      LOG.error("loadCommunities()"+" FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return result;    
  }  
  
  private boolean isCommunityNameFree(Connection conn, String communityName) throws SQLException {
    PreparedStatement pstmt = conn.prepareStatement("SELECT * FROM "+MASTER_TABLE+" WHERE name=? AND type=?");
    int i = 0;
    pstmt.setString(++i, communityName);
    pstmt.setInt(++i, MASTER_TYPE_COMMUNITY);
    ResultSet rs = pstmt.executeQuery();
    boolean result = !rs.next();
    LOG.debug("isCommunityNameFree("+communityName+"): "+result);
    return result;
  }
  
  private void grantOwnerRoles() throws WEPROMException {
    LOG.debug("grantOwnerRoles " + owner_id + " " + toString());
    Role role = RoleImpl.getInstance(Role.COMMUNITY_COORDINATOR);
    AccessRightsImpl._grantUserRole(owner_id, role, this);
    
    //role = RoleImpl.getInstance(Role.USER);
    //AccessRightsImpl._grantUserRole(owner_id, role, this);
    
    role = RoleImpl.getInstance(Role.COMMUNITY_PARTICIPANT);
    AccessRightsImpl._grantUserRole(owner_id, role, this);    
  }
  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Community[").append(id).append(',').append(name).append(']');
    return sb.toString();
  }
  
  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();    
    LOG.debug("SERIALIZE: " + toString());    
  }
  
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();    
    LOG.debug("DESERIALIZE: " + toString());    
  }

  int getDefaultStatus() {
    return PROJECT_STATUS_NOTSTARTED;
  }  
  
  static List getUserCommunities(User user) {
    String _method = "getUserCommunities("+user+") "; 
    LOG.debug(_method);
    int user_id = user.getID();
    List result = null;
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT DISTINCT ar.id FROM " + ACCESS_RIGHTS_TABLE + " AS ar INNER JOIN "+MASTER_TABLE+" ma ON ma.id=ar.id WHERE ar.user_id=? AND ma.type=?");
      pstmt.setInt(1, user_id);
      pstmt.setInt(2, MASTER_TYPE_COMMUNITY);
      rs = pstmt.executeQuery();
      result = new ArrayList();
      while (rs.next()) {
        int workspace_id = rs.getInt(1);
        LOG.debug(_method+"next community_id="+workspace_id);
        CommunityImpl c = CommunityImpl.getInstance(workspace_id);
        result.add(c);
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }   
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return result;
  }

  public void inviteUser(User user) throws WEPROMException {
    User caller = WepromRequestCycleUtil.currentUser();
    String _method = "inviteUser: " + caller + " invites " + user + " to " + toString(); 
    LOG.debug(_method);
    Permission p = PermissionImpl.getInstance(Permission.COMMUNITY_USER_EDIT_ROLES);
    
    boolean hasPermission = AccessRightsImpl._checkPermission(caller, this, p);
    LOG.debug(_method+" hasPermission="+hasPermission);
    
    if (hasPermission) {
      sendInvitationMessage(caller, user);
    } else {
      throw new AccessControlException(caller + " doesn't have permission to invite users into " + toString());
    }
    
  }
  
  private void sendInvitationMessage(User host, User guest) throws WEPROMException {
    MessageImpl msg = MessageImpl.createCommunityInvitationMessage(this, host, guest);
    msg.send();
  }  
}
