package com.fmi.weprom.impl.mysql;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Formatter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.Message;
import com.fmi.weprom.api.Role;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.WEPROMException;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.INVALID_ID;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_PROJECT;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_COMMUNITY;


public class MessageProcessor {
  
  private static final Log LOG = LogFactory.getLog(MessageProcessor.class);
  
  private static final String WORKSPACE_INVITATION_ACCEPTED_MSG_TEMPLATE = "User %s has %s the invitation to :\r\n %s : \"%s\"";
  private static final String TASK_ASSIGNMENT_ACCEPTED_MSG_TEMPLATE = "User %s has %s the assignment to:\r\n Task: \"%s\"\r\nin\r\n Project : \"%s\"";

  
  static void acceptWorkspaceInvitation(MessageImpl message, WorkspaceImpl workspace, Role role) throws WEPROMException {
    String _method = "acceptWorkspaceInvitation() "; 
    int receiver_id = message.getReceiverID();
    boolean transactionOK = false;
    Connection conn = null;
    try {
      conn = Util.currentSession().getConnection();
      LOG.debug(_method + "BEGIN TRANSACTION: ACCEPT_WORKSPACE_INVITATION");
      conn.setAutoCommit(false);
      
      LOG.debug(_method+"grant "+role.getName()+" role to user with id:"+receiver_id + " for " + workspace);

      AccessRightsImpl._grantUserRole(receiver_id, role, workspace);

      // set replied - if this fails it is bad
      message.setReplied(true);
      message.update();
      
      // Commit updates
      conn.commit();
      transactionOK = true;
      LOG.debug(_method + "END TRANSACTION: ACCEPT_WORKSPACE_INVITATION");
      
      // send message back to sender_id
      // this is not critical, so it is out of the transaction
      
      

    } catch (WEPROMException e) {
      LOG.error(_method+"FAILED to grant roles to user with id "+receiver_id + " for " + workspace, e);
      throw e;
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);      
      throw new WEPROMException("Faild to accept workspace invitation");
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug(_method + "ROLLBACK TRANSACTION: ACCEPT_WORKSPACE_INVITATION");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}  
    }    
  }
  
  static void acceptProjectInvitation(MessageImpl message) throws WEPROMException {
    String _method = "acceptProjectInvitation("+message+") "; 
    LOG.debug(_method);
    int projectID = message.extractProjectID();
    int receiver_id = message.getReceiverID();
    if (projectID!=INVALID_ID) {
      ProjectImpl project = ProjectImpl.getInstance(projectID);
      if (project!=null) {
        
        acceptWorkspaceInvitation(message, project, RoleImpl.getInstance(Role.PROJECT_PARTICIPANT));
        
        // send message back to sender_id
        // this is not critical, so it is out of the transaction
        try {
          MessageImpl reportMsg = createWorkspaceInvitationReportMessage(MASTER_TYPE_PROJECT, project.getName(), message, true);
          reportMsg.send();
        } catch (Exception e) {
          LOG.debug(_method+"FAILED to send report message", e);
        }
      } else {
        LOG.debug(_method+"FAILED: project with id "+projectID+" does not exist");
      }
    } else {
      LOG.debug(_method+"FAILED: invalid project id");
    }    
  }

  static void acceptCommunityInvitation(MessageImpl message) throws WEPROMException {
    String _method = "acceptCommunityInvitation("+message+") "; 
    LOG.debug(_method);
    int communityID = message.extractCommunityID();
    if (communityID!=INVALID_ID) {
      CommunityImpl community = CommunityImpl.getInstance(communityID);
      if (community!=null) {
        acceptWorkspaceInvitation(message, community, RoleImpl.getInstance(Role.COMMUNITY_PARTICIPANT));

        // send message back to sender_id
        // this is not critical, so it is out of the transaction
        try {
          MessageImpl reportMsg = createWorkspaceInvitationReportMessage(MASTER_TYPE_COMMUNITY, community.getName(), message, true);
          reportMsg.send();
        } catch (Exception e) {
          LOG.debug(_method+"FAILED to send report message", e);
        }
      } else {
        LOG.debug(_method+"FAILED: community with id "+communityID+" does not exist");
      }
    } else {
      LOG.debug(_method+"FAILED: invalid community id");
    }      
  }

  static void acceptTaskAssignment(MessageImpl message) throws WEPROMException {
    String _method = "acceptTaskAssignment("+message+") "; 
    LOG.debug(_method);
   
    int taskID = message.extractTaskID();
    if (taskID!=INVALID_ID) {
      TaskImpl task = TaskImpl.getInstance(taskID);
      if (task!=null) {
        
        boolean transactionOK = false;
        Connection conn = null;
        try {
          conn = Util.currentSession().getConnection();
          LOG.debug(_method + "BEGIN TRANSACTION: ACCEPT_TASK_ASSIGNMENT");
          conn.setAutoCommit(false);
          
          // delete this message or set replied
          message.setReplied(true);
          message.update();
          
          task.setAssigned(true);
          task.update(); // note that this commits!
          
          // Commit updates
          if (!conn.getAutoCommit()) {
            conn.commit();
          }
          transactionOK = true;
          LOG.debug(_method + "END TRANSACTION: ACCEPT_TASK_ASSIGNMENT");          
          
          // send info message back to sender_id
          // this is not critical, so it is out of the transaction
          sendTaskAssignmentReportMessage(message, task, true);
          
        } catch (WEPROMException e) {
          LOG.error(_method+"FAILED to assign task " + task, e);
          throw e;
        } catch (SQLException e) {
          LOG.error(_method+"FAILED", e);      
          throw new WEPROMException("Faild to accept task assignment");
        } finally {
          if (!transactionOK) {
            try {
              LOG.debug(_method + "ROLLBACK TRANSACTION: ACCEPT_PROJECT_INVITATION");
              conn.rollback();
            } catch (SQLException ignore) {}
          }
          try {
            conn.setAutoCommit(true);
          } catch (SQLException ignore) {}            
        }
      } else {
        LOG.debug(_method+"FAILED: task with id "+taskID+" does not exist");
      }
    } else {
      LOG.debug(_method+"FAILED: invalid task id");
    }
  }

  static void declineProjectInvitation(MessageImpl message) {
    String _method = "declineProjectInvitation("+message+") "; 
    LOG.debug(_method);
    try {
      message.setReplied(true);
      message.update();
      
      // send report message back
      int projectID = message.extractProjectID();
      if (projectID!=INVALID_ID) {
        ProjectImpl project = ProjectImpl.getInstance(projectID);
        if (project!=null) {      
          try {
            MessageImpl reportMsg = createWorkspaceInvitationReportMessage(MASTER_TYPE_PROJECT, project.getName(), message, false);
            reportMsg.send();
          } catch (Exception e) {
            LOG.debug(_method+"FAILED to send report message", e);
          }             
        } else {
          LOG.debug(_method+"FAILED to send report message: "+"project("+projectID+") could not be loaded");
        }
      } else {
        LOG.debug(_method+"FAILED to send report message: "+"invalid project id");
      }
    } catch (WEPROMException e) {
      LOG.error(_method+"FAILED to decline project invitation", e);      
    }    
  }

  static void declineCommunityInvitation(MessageImpl message) {
    String _method = "declineCommunityInvitation("+message+") "; 
    LOG.debug(_method);
    try {
      message.setReplied(true);
      message.update();
      
      // send report message back
      int communityID = message.extractCommunityID();
      if (communityID!=INVALID_ID) {
        CommunityImpl community = CommunityImpl.getInstance(communityID);
        if (community!=null) {      
          try {
            MessageImpl reportMsg = createWorkspaceInvitationReportMessage(MASTER_TYPE_COMMUNITY, community.getName(), message, false);
            reportMsg.send();
          } catch (Exception e) {
            LOG.debug(_method+"FAILED to send report message", e);
          }             
        } else {
          LOG.debug(_method+"FAILED to send report message: "+"community("+communityID+") could not be loaded");
        }
      } else {
        LOG.debug(_method+"FAILED to send report message: "+"invalid community id");
      }      
    } catch (WEPROMException e) {
      LOG.error(_method+"FAILED to decline community invitation", e);      
    }   
  }

  static void declineTaskAssignment(MessageImpl message) throws WEPROMException {
    String _method = "declineTaskAssignment("+message+") "; 
    LOG.debug(_method);
    
    int taskID = message.extractTaskID();
    if (taskID!=INVALID_ID) {
      TaskImpl task = TaskImpl.getInstance(taskID);
      if (task!=null) {
        
        boolean transactionOK = false;
        Connection conn = null;        
        try {
          conn = Util.currentSession().getConnection();
          LOG.debug(_method + "BEGIN TRANSACTION: DECLINE_TASK_ASSIGNMENT");
          conn.setAutoCommit(false);
          
          // delete this message or set replied
          message.setReplied(true);
          message.update();
          
          task.setAssigned(false);
          task.setProcessor(null);
          task.update(); // note that this commits!

          // Commit updates
          if (!conn.getAutoCommit()) {
            conn.commit();
          }
          transactionOK = true;
          LOG.debug(_method + "END TRANSACTION: DECLINE_TASK_ASSIGNMENT");  
          
          // send info message back to sender_id
          // this is not critical, so it is out of the transaction 
          sendTaskAssignmentReportMessage(message, task, false);

        } catch (WEPROMException e) {
          LOG.error(_method+"FAILED to decline task " + task, e);
          throw e;
        } catch (SQLException e) {
          LOG.error(_method+"FAILED", e);      
          throw new WEPROMException("Faild to decline task assignment");
        } finally {
          if (!transactionOK) {
            try {
              LOG.debug(_method + "ROLLBACK TRANSACTION: DECLINE_TASK_ASSIGNMENT");
              conn.rollback();
            } catch (SQLException ignore) {}
          }
          try {
            conn.setAutoCommit(true);
          } catch (SQLException ignore) {}           
        }
      } else {
        LOG.debug(_method+"FAILED: task with id "+taskID+" does not exist");
      }
    } else {
      LOG.debug(_method+"FAILED: invalid task id");
    }   
  }
  
  private static void sendTaskAssignmentReportMessage(MessageImpl message, TaskImpl task, boolean accepted) {
    int projectID = message.extractProjectID();
    if (projectID!=INVALID_ID) {
      ProjectImpl project = ProjectImpl.getInstance(projectID);
      if (project!=null) {
        MessageImpl reportMsg = createTaskAssignmentReportMessage(task, project.getName(), message, accepted);
        try {
          reportMsg.send();
        } catch (Exception e) {
          LOG.debug("sendTaskAssignmentReportMessage() FAILED to report to message: " + message, e);
        }
      }
    }
  }
  
  private static MessageImpl createTaskAssignmentReportMessage(TaskImpl task, String workspaceName, MessageImpl message, boolean accepted) {
    User guest = message.getReceiver();
    User host = message.getSender();
    LOG.debug("createTaskAssignmentReportMessage("+workspaceName+", "+guest+" >> "+host+")");
    MessageImpl msg = new MessageImpl((UserWorkspaceImpl) guest.getUserWorkspace());
    String action = null;
    StringBuilder sb = new StringBuilder();
    if (accepted) {
      sb.append("[Accepted]");
      action = "accepted";
    } else {
      sb.append("[Declined]");
      action = "declined";
    }
    sb.append(' ').append(message.getSubject());
    msg.setSubject(sb.toString());
    
    sb.setLength(0); // clear
    Formatter formatter = new Formatter(sb);
    formatter.format(TASK_ASSIGNMENT_ACCEPTED_MSG_TEMPLATE, guest.getNick(), action, task.getSubject(),workspaceName);
    formatter.flush();
    
    msg.setBody(sb.toString());
    msg.setType(Message.MESSAGE_TYPE_TEXT);
    msg.setReceiver(host);
    return msg;
  } 
  
  private static MessageImpl createWorkspaceInvitationReportMessage(int type, String workspaceName, MessageImpl message, boolean accepted) {
    User guest = message.getReceiver();
    User host = message.getSender();
    LOG.debug("createWorkspaceInvitationReportMessage("+workspaceName+", "+guest+" >> "+host+")");
    MessageImpl msg = new MessageImpl((UserWorkspaceImpl) guest.getUserWorkspace());
    String action = null;
    
    StringBuilder sb = new StringBuilder();
    if (accepted) {
      sb.append("[Accepted]");
      action = "accepted";
    } else {
      sb.append("[Declined]");
      action = "declined";
    }
    String workspaceTypeName = getWorkspaceTypeName(type);
    sb.append(' ').append(workspaceTypeName).append(" Invitation : ").append(workspaceName);
    msg.setSubject(sb.toString());
    
    sb.setLength(0); // clear
    Formatter formatter = new Formatter(sb);
    
    formatter.format(WORKSPACE_INVITATION_ACCEPTED_MSG_TEMPLATE, guest.getNick(), action, workspaceTypeName, workspaceName);
    formatter.flush();
    
    msg.setBody(sb.toString());
    msg.setType(Message.MESSAGE_TYPE_TEXT);
    msg.setReceiver(host);
    return msg;
  }  
  
  private static String getWorkspaceTypeName(int type) {
    switch (type) {
      case MASTER_TYPE_PROJECT:
        return "Project";
        
      case MASTER_TYPE_COMMUNITY:
        return "Community";
        
      default:
        return "UNKNOWN";
    }
  }


}
