package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_RIGHTS_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_TYPE_SHARED;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_PROJECT;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.UNSAVED_ID;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.AccessControlException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.AccessRights;
import com.fmi.weprom.api.Calendar;
import com.fmi.weprom.api.File;
import com.fmi.weprom.api.Permission;
import com.fmi.weprom.api.Project;
import com.fmi.weprom.api.Role;
import com.fmi.weprom.api.Task;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.WEPROMException;
import com.fmi.weprom.util.WepromRequestCycleUtil;

public class ProjectImpl extends WorkspaceImpl implements Project {
  
  private static final Log LOG = LogFactory.getLog(ProjectImpl.class);
  
  private static final String CACHE_NAME = "ProjectCache";
  private static final int INVALID_USER = -1;
  
  protected transient User owner;
  protected String name;
  protected int owner_id = INVALID_USER;
  protected transient CalendarImpl calendar;
  protected transient WorkspaceInfo info;

  
  
  ProjectImpl(User user) {
    this(user, null, UNSAVED_ID);
  }

  private ProjectImpl(int owner_id, String name, int id) {
    super();
    this.owner_id = owner_id;
    this.name = name;
    this.id = id;  
  }

  private ProjectImpl(User owner, String name, int id) {
    super();
    this.owner = owner;
    this.owner_id = owner.getID();
    this.name = name;
    this.id = id;  
  }
  
  public Calendar getCalendar() {
    if (calendar==null) {
      calendar = new CalendarImpl(this);
    }
    return calendar;
  }
  
  public Task createTask(User taskOwner) {
    LOG.debug("createTask");
    Permission p = PermissionImpl.getInstance(Permission.CREATE_PROJECT_TASK);
    
    boolean hasPermission = AccessRightsImpl._checkPermission(taskOwner, this, p);
    LOG.debug("createTask user "+taskOwner.getNick()+" hasPermission="+hasPermission);
    Task task = null;
    if (hasPermission) {
      task = new TaskImpl(this, taskOwner); 
    }
    return task;
  }

  public List getTasks() {
    return TaskImpl.loadProjectTasks(getID());
  }

  public Task getTask(int id) {
    return TaskImpl.getInstance(id);
  }

  public List getUsers() {
    return WorkspaceImpl.getUsers(LOG, id);
    /*
    LOG.debug("getUsers() ");
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Project does not exists");
    }    
    List result = null;
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT DISTINCT user_id FROM " + ACCESS_RIGHTS_TABLE + " WHERE id=?");
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      
      result = new ArrayList();
      while (rs.next()) {
        int user_id = rs.getInt(1);
        LOG.debug("getUsers() "+"user_id="+user_id);
        UserImpl user = UserImpl.getInstance(user_id);
        result.add(user);
      }
    } catch (SQLException e) {
      LOG.error("getUsers() " + "FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }   
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return result;
    */
  }

  private boolean isProjectNameFree(Connection conn, String projectName) throws SQLException {
    PreparedStatement pstmt = conn.prepareStatement("SELECT * FROM "+MASTER_TABLE+" WHERE name=? AND type=?");
    int i = 0;
    pstmt.setString(++i, projectName);
    pstmt.setInt(++i, MASTER_TYPE_PROJECT);
    ResultSet rs = pstmt.executeQuery();
    boolean result = !rs.next();
    LOG.debug("isProjectNameFree("+projectName+"): "+result);
    return result;
  }
  
  public void setName(String name) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Project already exists - name could not be changed");
    }
    this.name = name;    
  }
  
  public void update() throws WEPROMException {
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Project does not exists");
    }
    
    // check permission
    User caller = WepromRequestCycleUtil.currentUser();
    Permission p = PermissionImpl.getInstance(Permission.UPDATE_PROJECT);
    boolean hasPermission = AccessRightsImpl._checkPermission(caller, this, p);
    LOG.debug("update() checkPermission for " + caller + " : " + hasPermission);
    if (!hasPermission) return;
    
    try {
      getInfo().update();
    } catch (SQLException e) {
      throw new WEPROMException("FAILED to update project: "+toString(), e);
    }
  }
  
  public void save() throws WEPROMException {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Project already exists");
    }
    if (name==null) {
      throw new IllegalStateException("Project does not have name");
    }
    LOG.debug("save(...) id="+id + " user.id="+owner_id);

    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    boolean transactionOK = false;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement();
      
      LOG.debug("save() BEGIN TRANSACTION: CREATE_PROJECT");
      // Disable auto commit
      conn.setAutoCommit(false);
      
      // check if the project name is not alreay in use
      if (isProjectNameFree(conn, name)) {      
        StringBuffer sb = new StringBuffer();
        sb.append("INSERT INTO").append(' ').append(MASTER_TABLE).append(' ');
        sb.append('(');
        sb.append("type").append(',');
        sb.append("owner_id").append(',');
        sb.append("access_type").append(',');
        sb.append("name").append(',');
        sb.append("creation_date");
        sb.append(')');
        sb.append(' ').append("VALUES").append('(');
        sb.append('\'').append(MASTER_TYPE_PROJECT).append('\'').append(',');
        sb.append('\'').append(owner_id).append('\'').append(',');
        sb.append('\'').append(ACCESS_TYPE_SHARED).append('\'').append(',');
        sb.append('\'').append(name).append('\'').append(',');
        sb.append("NOW()");
        sb.append(')');
        String sql = sb.toString();
        LOG.debug("save() sql="+sql);
        
        stmt.executeUpdate(sql, Statement.RETURN_GENERATED_KEYS);
        
        WorkspaceInfo info = getInfo();
        
        // get autogenerated project id
        rs = stmt.getGeneratedKeys();
        if (rs.next()) {
          int projectID = rs.getInt(1);
          this.id = projectID; 
          LOG.debug("save() generated id="+id);
          
          // grant Role.PROJECT_COORDINATOR to the user
          Role role = RoleImpl.getInstance(Role.PROJECT_COORDINATOR);
          AccessRightsImpl._grantUserRole(owner, role, this);
          
          //role = RoleImpl.getInstance(Role.USER);
          //AccessRightsImpl._grantUserRole(owner, role, this);
          
          role = RoleImpl.getInstance(Role.PROJECT_PARTICIPANT);
          AccessRightsImpl._grantUserRole(owner, role, this);
          
          // insert into workspaces
          //WorkspaceInfo info = createInfo();
          info.save();
          
          // Commit updates
          conn.commit();
          transactionOK = true;
          LOG.debug("save() END TRANSACTION: CREATE_PROJECT");          
          cache();
        } else {
          LOG.debug("save() FAILED for "+toString());
          throw new WEPROMException("Faild to create project");
        }       
      } else {
        throw new WEPROMException("Faild to create project - name is already in use");
      }
    } catch (SQLException e) {
      LOG.error("save() FAILED", e);
      throw new WEPROMException("Faild to create project! Project name maybe already in use.");
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("save() ROLLBACK TRANSACTION: CREATE_PROJECT");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}      
      /*
      if (conn!=null) {
        try {
          conn.close();
        } catch (SQLException ignore) {}
      }
      */
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    
  }

  public void delete() {
    // TODO Auto-generated method stub
    // delete all releated tasks and reminders
    //getTasks();
    uncache();
  }

  public String getName() {
    return name;
  }

  private void uncache() {
    LOG.debug("uncache: " + toString());
    Map userCache = Util.currentSession().getCache(CACHE_NAME);
    userCache.remove(new Integer(id));
  }
  
  private void cache() {
    LOG.debug("cache: " + toString());
    Map userCache = Util.currentSession().getCache(CACHE_NAME);
    userCache.put(new Integer(id), this);
  }
  
  static ProjectImpl getCached(int id) {
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    Integer _id = new Integer(id);
    ProjectImpl project = (ProjectImpl) cache.get(_id);
    return project;
  }

  static ProjectImpl getInstance(int id) {
    ProjectImpl project = getCached(id);
    if (project!=null) {
      return project;
    }
    return load(id);
  }
  
  private static ProjectImpl load(int id) {
    String _method = "load(" + id + ") "; 
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT name, owner_id FROM "+MASTER_TABLE+ " WHERE id=?"); 
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      while (rs.next()) {
        int i = 0;
        String name = rs.getString(++i);
        int owner_id = rs.getInt(++i);
        ProjectImpl p = ProjectImpl.getInstance(owner_id, name, id);
        return p;
      }
      
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;    
  }
  
  static List loadProjects() {
    LOG.debug("loadProjects()");
    List result = new ArrayList();
    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement(); 
      rs = stmt.executeQuery("SELECT id, name, owner_id FROM "+MASTER_TABLE+ " where type='"+MASTER_TYPE_PROJECT+"'");
      while (rs.next()) {
        int id = rs.getInt(1);
        String name = rs.getString(2);
        int owner_id = rs.getInt(3);
        //User owner = UserImpl.getUser(weprom, owner_id);
        ProjectImpl p = ProjectImpl.getInstance(owner_id, name, id);
        result.add(p);
      }
      
    } catch (SQLException e) {
      LOG.error("loadProjects() FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return result;    
  }
  
  static ProjectImpl getInstance(int owner_id, String name, int id) {
    ProjectImpl project = getCached(id);
    if (project==null) {
      project = new ProjectImpl(owner_id, name, id);
      LOG.debug("getInstance() "+"create new : " + project);
      project.cache();
    } else {
      LOG.debug("getInstance() "+"load cached : " + project);
    }
    return project;    
  }

  public User getOwner() {
    if (owner==null && owner_id!=INVALID_USER) {
      // lazy load the owner
      owner  = UserImpl.getInstance(owner_id);
    }
    return owner;
  }
  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Project[").append(id).append(',').append(name).append(']');
    return sb.toString();
  }
  
  private WorkspaceInfo createInfo() {
    return new WorkspaceInfo(this);
  }
  
  private WorkspaceInfo getInfo() {
    if (info==null) {
      if (id==UNSAVED_ID) {
        info = createInfo();
      } else {
        info = WorkspaceInfo.load(this);
      }
    }
    return info;
  }
  
  public int getCompleted() {
    return getInfo().getCompleted();
  }


  public void setCompleted(int completed) {
    getInfo().setCompleted(completed);
  }


  public String getDescription() {
    return getInfo().getDescription();
  }


  public void setDescription(String description) {
    getInfo().setDescription(description);
  }


  public Date getEndDate() {
    return getInfo().getEndDate();
  }


  public void setEndDate(Date endDate) {
    getInfo().setEndDate(endDate);
  }


  public Date getStartDate() {
    return getInfo().getStartDate();
  }


  public void setStartDate(Date startDate) {
    getInfo().setStartDate(startDate);
  }


  public int getStatus() {
    return getInfo().getStatus();
  }


  public void setStatus(int status) {
    getInfo().setStatus(status);
  }  
  
  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();    
    LOG.debug("SERIALIZE: " + toString());    
  }
  
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();    
    LOG.debug("DESERIALIZE: " + toString());    
  }

  int getType() {
    return MASTER_TYPE_PROJECT;
  }

  public void inviteUser(User user) throws WEPROMException {
    User caller = WepromRequestCycleUtil.currentUser();
    String _method = "inviteUser: " + caller + " invites " + user + " to " + toString(); 
    LOG.debug(_method);
    Permission p = PermissionImpl.getInstance(Permission.PROJECT_USER_EDIT_ROLES);
    
    boolean hasPermission = AccessRightsImpl._checkPermission(caller, this, p);
    LOG.debug(_method+" hasPermission="+hasPermission);
    
    if (hasPermission) {
      sendInvitationMessage(caller, user);
    } else {
      throw new AccessControlException(caller + " doesn't have permission to invite users into " + toString());
    }
  }  
  
  private void sendInvitationMessage(User host, User guest) throws WEPROMException {
    MessageImpl msg = MessageImpl.createProjectInvitationMessage(this, host, guest);
    msg.send();
  }

  int getDefaultStatus() {
    return PROJECT_STATUS_NOTSTARTED;
  }
  
  static List getUserProjects(User user) {
    String _method = "getUserProjects("+user+") ";
    LOG.debug(_method);
    int user_id = user.getID();
    List result = null;
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT DISTINCT ar.id FROM " + ACCESS_RIGHTS_TABLE + " AS ar INNER JOIN "+MASTER_TABLE+" ma ON ma.id=ar.id WHERE ar.user_id=? AND ma.type=?");
      
      pstmt.setInt(1, user_id);
      pstmt.setInt(2, MASTER_TYPE_PROJECT);
      rs = pstmt.executeQuery();
      
      result = new ArrayList();
      while (rs.next()) {
        int workspace_id = rs.getInt(1);
        LOG.debug(_method+"next workspace_id="+workspace_id);
        ProjectImpl p = ProjectImpl.getInstance(workspace_id);
        result.add(p);
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }   
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return result;
  }


}
