package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.REMINDERS_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.UNSAVED_ID;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.Reminder;
import com.fmi.weprom.api.WEPROMException;
import com.fmi.weprom.util.WepromRequestCycleUtil;

public class ReminderImpl implements Reminder, Serializable {
  
  private static final Log LOG = LogFactory.getLog(ReminderImpl.class);
  private static final String CACHE_NAME = "ReminderCache";
  private static final int DEFAULT_STATUS = STATUS_ON;
  
  protected transient CalendarImpl calendar;
  protected int id;
  protected Date date;
  protected String subject;
  protected String description;
  protected int status = DEFAULT_STATUS;
  
  
  public ReminderImpl(CalendarImpl calendar) {
    this.calendar = calendar;
    this.id = UNSAVED_ID;
  }
  
  private ReminderImpl(int id, CalendarImpl calendar, Date date, String subject, String description, int status) {
    this.calendar = calendar;
    this.id = id;
    this.date = date;
    this.subject = subject;
    this.description = description;
    this.status = status;
  }

  public int getID() {
    return id;
  }

  public void save() throws WEPROMException {
    LOG.debug("save()");    
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Reminder already exists");
    }
    save(true);
  }
  
  private void save(boolean unsaved) throws WEPROMException {
    if (date==null) {
      throw new IllegalStateException("Reminder does not have a memo date");
    }    
    final String _method = "save("+unsaved+") "; 
    LOG.debug(_method);

    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    int i = 0;
    try {
      conn = Util.currentSession().getConnection();
      if (unsaved) {
        pstmt = conn.prepareStatement("INSERT INTO "+REMINDERS_TABLE+" (workspace_id,date,subject,description,status) VALUES(?,?,?,?,?)", 
                                  PreparedStatement.RETURN_GENERATED_KEYS);
      } else {
        pstmt = conn.prepareStatement("UPDATE reminders SET date=?,subject=?,description=?,status=? WHERE id=?");        
      }
      if (unsaved) pstmt.setInt(++i, calendar.getWorkspaceID());
      pstmt.setTimestamp(++i, new Timestamp(date.getTime()));
      pstmt.setString(++i, subject);
      pstmt.setString(++i, description);
      pstmt.setInt(++i, status);
      if (!unsaved) pstmt.setInt(++i, id);
      
      pstmt.executeUpdate();

      if (unsaved) {
        // get autogenerated project id
        rs = pstmt.getGeneratedKeys();
        
        if (rs.next()) {
          this.id = rs.getInt(1);
          LOG.debug(_method + "generated id="+id);
          cache();
        } else {
            LOG.debug(_method + "FAILED for "+toString());
            throw new WEPROMException("Faild to create reminder");
        }
      }
    } catch (SQLException e) {
      LOG.error(_method + "FAILED", e);      
      throw new WEPROMException("Faild to create reminder!");
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }        
  }  

  public void delete() throws WEPROMException {
    LOG.debug("delete() ");
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Reminder is unsaved and could not be deleted");
    }
    
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("DELETE FROM "+REMINDERS_TABLE+" WHERE id=?");
      pstmt.setInt(1, id);
      pstmt.executeUpdate();
      id = UNSAVED_ID;
    } catch (SQLException e) {
      LOG.error("delete() " + "FAILED", e);      
      throw new WEPROMException("Faild to delete reminder");
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }     
    
    uncache();    
  }

  public void update() throws WEPROMException {
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Reminder is unsaved and could not be updated");
    }
    LOG.debug("update()");
    save(false);
  }


  static List loadReminders(CalendarImpl calendar) {
    return loadReminders(calendar, null, null);
  }
  
  static List loadReminders(CalendarImpl calendar, Date startDate, Date endDate) {
    int workspace_id = calendar.getWorkspaceID();
    LOG.debug("loadReminders() for workspace:"+workspace_id);
    
    List result = null;
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      if (startDate==null && endDate==null) {
        // all dates
        pstmt = conn.prepareStatement("SELECT id,date,subject,description,status FROM "+REMINDERS_TABLE+" WHERE workspace_id=?");
        pstmt.setInt(1, workspace_id);
      } else if (startDate!=null && endDate==null) {
        // all in [startDate, ...]
        pstmt = conn.prepareStatement("SELECT id,date,subject,description,status FROM "+REMINDERS_TABLE+" WHERE workspace_id=?" +
                                      " AND date>=?");
        pstmt.setInt(1, workspace_id);
        pstmt.setTimestamp(2, new Timestamp(startDate.getTime()));
      } else if (startDate==null && endDate!=null) {
        // all in [..., endDate]
        pstmt = conn.prepareStatement("SELECT id,date,subject,description,status FROM "+REMINDERS_TABLE+" WHERE workspace_id=?" +
        " AND date<=?");
        pstmt.setInt(1, workspace_id);
        pstmt.setTimestamp(2, new Timestamp(endDate.getTime()));        
      } else {
        // all in [startDate, endDate]
        pstmt = conn.prepareStatement("SELECT id,date,subject,description,status FROM "+REMINDERS_TABLE+" WHERE workspace_id=?" +
            " AND date>=? AND date<=?");
        pstmt.setInt(1, workspace_id);
        pstmt.setTimestamp(2, new Timestamp(startDate.getTime()));
        pstmt.setTimestamp(3, new Timestamp(endDate.getTime()));
      }
      
      rs = pstmt.executeQuery();
      
      result = new ArrayList();
      while (rs.next()) {
        int id = rs.getInt(1);
        Timestamp date_timestamp = rs.getTimestamp(2);
        String subject = rs.getString(3);
        String description = rs.getString(4);
        int status = rs.getInt(5);
        
        Date date = null;
        if (date_timestamp!=null) {
          date = new Date(date_timestamp.getTime()); 
        }
        
        ReminderImpl reminder = getInstance(id, calendar, date, subject, description, status);
        LOG.debug("loadReminders() add reminder="+reminder);
        result.add(reminder);
      }
      
    } catch (SQLException e) {
      LOG.error("loadReminders() " + "FAILED", e);      
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }         
    return result;
  }
  
  private static ReminderImpl load(CalendarImpl calendar, int id) {
    String _method = "loadReminder("+id+") ";
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT date,subject,description,status FROM "+
          REMINDERS_TABLE+" WHERE id=?"); 
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      if (rs.next()) {
        int i = 0;
        Timestamp date_timestamp = rs.getTimestamp(++i);
        String subject = rs.getString(++i);
        String description = rs.getString(++i);
        int status = rs.getInt(++i);

        Date date = null;
        if (date_timestamp!=null) {
          date = new Date(date_timestamp.getTime()); 
        }
        ReminderImpl reminder = getInstance(id, calendar, date, subject, description, status);
        return reminder;
      }
    } catch (SQLException e) {
      LOG.error(_method + "FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;    
  }
  
  private static ReminderImpl getCached(int id) {    
    //LOG.debug("getCached id=" + id);
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    return (ReminderImpl) cache.get(new Integer(id));
  }
  
  static ReminderImpl getInstance(int id, CalendarImpl calendar, Date date, String subject, String description, int status) {
    ReminderImpl reminder = getCached(id);
    if (reminder==null) {      
      LOG.debug("getInstance(...) create");
      reminder = new ReminderImpl(id, calendar, date, subject, description, status);
      reminder.cache();
    } else {
      LOG.debug("getInstance(...) loaded from cache : " + reminder);
      reminder.setDate(date);
      reminder.setSubject(subject);
      reminder.setDescription(description);
      reminder.setStatus(status);
    }
    return reminder;
  }
  
  public static ReminderImpl getInstance(CalendarImpl calendar, int id){
    ReminderImpl reminder = getCached(id);
    if (reminder==null) {
      reminder = load(calendar, id);
    }
    return reminder;
  }
  
  public Date getDate() {
    return date;
  }

  public String getSubject() {
    return subject;
  }

  public String getDescription() {
    return description;
  }

  public int getStatus() {
    return status;
  }

  public void setDate(Date date) {
    this.date = date;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public void setStatus(int status) {
    this.status = status;
  }

  public void setSubject(String subject) {
    this.subject = subject;
  }
  
  private void uncache() {
    LOG.debug("uncache: " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.remove(this);
  }
  
  private void cache() {
    LOG.debug("cache : " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.put(new Integer(id), this);
  }  

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Reminder[").append(subject).append(',').append(date).append(']');
    return sb.toString();
  }

  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();    
    LOG.debug("SERIALIZE: " + toString());    
  }
  
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();    
    LOG.debug("DESERIALIZE: " + toString());    
  }
}
