package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_RIGHTS_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.USERS_TABLE;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.AccessRights;
import com.fmi.weprom.api.Permission;
import com.fmi.weprom.api.Role;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.UserWorkspace;
import com.fmi.weprom.api.WEPROMException;
import com.fmi.weprom.api.Workspace;
import com.fmi.weprom.util.WepromRequestCycleUtil;

public class AccessRightsImpl implements AccessRights {
  
  private static final Log LOG = LogFactory.getLog(AccessRightsImpl.class);
  private static final String ROLE_CACHE = "RoleCache";
  private static final String ROLE_ID_CACHE = "RoleIDCache";
  
  public void grantUserRole(User user, Role role, Workspace w) throws WEPROMException {
    _grantUserRole(user, role, w);
  }

  static void _grantUserRole(User user, Role role, Workspace w) throws WEPROMException {
    LOG.debug("grantUserRole: " + user + "; "+role+"; " + w);   
    _grantUserRole(user.getID(), role, w);
  }
  
  // note that this method participates in other transactions, so it should not commit!
  static void _grantUserRole(int user_id, Role role, Workspace w) throws WEPROMException {
    LOG.debug("grantUserRole: user:" + user_id + "; "+role+"; " + w);    
    Connection conn = null;
    Statement stmt = null;
    SQLException exc = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement();
      //grantUserRole(stmt, user, role, w);
      StringBuffer sb = new StringBuffer();
      sb.append("INSERT INTO " + ACCESS_RIGHTS_TABLE + " (id,user_id,role_id) VALUES(");
      sb.append(w.getID()).append(',');
      sb.append(user_id).append(',');
      sb.append(role.getID());
      sb.append(')');
      String sql = sb.toString();
      //LOG.debug("grantUserRole sql="+sql);
      
      stmt.executeUpdate(sql);      
    } catch (SQLException e) {
      LOG.error("grantUserRole() FAILED", e);
      exc = e;
    } finally {
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }
      if (exc!=null) {
        throw new WEPROMException("grantUserRole failed", exc);
      }
    }    
  }
  
  public void grantUserRoles(User user, List roles, Workspace w) throws WEPROMException {
    _grantUserRoles(user, roles, w);
  }
  
  public static void _grantUserRoles(User user, List roles, Workspace w) throws WEPROMException {
    Connection conn = null;
    PreparedStatement pstmt = null;
    SQLException exc = null;
    boolean transactionOK = false;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("INSERT INTO " + ACCESS_RIGHTS_TABLE + " (id,user_id,role_id) VALUES(?,?,?)");
      
      LOG.debug("grantUserRoles() BEGIN TRANSACTION: GRANT_USER_ROLES");
      conn.setAutoCommit(false);
      
      int user_id = user.getID();
      int id = w.getID();
      int size = roles.size();
      for (int i = 0; i < roles.size(); i++) {
        Role role = (Role) roles.get(i);
        pstmt.setInt(1, id);
        pstmt.setInt(2, user_id);
        pstmt.setInt(3, role.getID());
        pstmt.executeUpdate();
        LOG.debug("grantUserRoles() user:"+user+" role:"+role+" workspace:"+w);
      }
      
      conn.commit();
      transactionOK = true;
      LOG.debug("grantUserRoles() END TRANSACTION: GRANT_USER_ROLES");
      conn.setAutoCommit(true);
      
    } catch (SQLException e) {
      LOG.error("grantUserRoles() FAILED", e);
      exc = e;
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("save() ROLLBACK TRANSACTION: GRANT_USER_ROLES");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}      
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
      if (exc!=null) {
        throw new WEPROMException("grantUserRole failed", exc);
      }
    }        
  }
  
  public void deleteUserRoles(User user, Workspace w) {
    LOG.debug("deleteUserRoles("+user.getNick()+","+w.getID()+")");
    Connection conn = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("DELETE FROM " + ACCESS_RIGHTS_TABLE + " WHERE id=? AND user_id=?");
      
      pstmt.setInt(1, w.getID());
      pstmt.setInt(2, user.getID());
      pstmt.executeUpdate();
      
    } catch (SQLException e) {
      LOG.error("deleteUserRoles() FAILED", e);
    } finally {
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }        
  }
  
  public boolean checkPermission(User user, Workspace w, Permission p) {
    return _checkPermission(user, w, p);
  }
  
  static boolean _checkPermission(User user, Workspace w, Permission p) {
    return _checkPermission(user, w.getID(), p);
  }
  
  static boolean _checkPermission(User user, int workspaceID, Permission p) {
    // alternative way is to use a direct query:
    // SELECT * FROM access_rights WHERE user_id=3 AND 
    //                                   role_id IN (select role_id from role_permissions where 
    //                                               permission_id IN (select id from permissions where name='reminder.view'))
    String _method = "checkPermission("+user.getNick()+","+workspaceID+","+p+") "; 
    LOG.debug(_method);
    List roles = _getUserRoles(user, workspaceID);
    int size = roles.size();
    for (int i=0; i<size; i++) {
      Role role = (Role) roles.get(i);
      List permissions = role.getPermissions();
      int p_size = permissions.size();
      for (int j = 0; j < p_size; j++) {
        PermissionImpl permission = (PermissionImpl) permissions.get(j);
        boolean implies = permission.implies(p);
        LOG.debug(permission + " implies " + p + " : "+implies);
        if (implies) {
          LOG.debug(_method + ": true");
          return true;
        }
      }
    } 
    LOG.debug(_method + ": false");
    return false;
  }
  
  public List getUserRoles(User user, Workspace w) {
    return _getUserRoles(user, w);
  }
  
  static List _getUserRoles(User user, Workspace w) {
    return _getUserRoles(user, w.getID());
  }
  
  static List _getUserRoles(User user, int workspaceID) {
    LOG.debug("getUserRoles("+user+", "+workspaceID+")");
    
    List result = new ArrayList();
    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement();
      
      StringBuilder sb = new StringBuilder();
      sb.append("SELECT role_id FROM ").append(ACCESS_RIGHTS_TABLE).append(" WHERE ");
      sb.append("user_id='").append(user.getID()).append('\'');
      sb.append(" AND ");
      sb.append("id='").append(workspaceID).append('\'');
      
      String sql = sb.toString();
      //LOG.debug("getUserRoles sql="+sql);
      
      rs = stmt.executeQuery(sql);
      while (rs.next()) {
        int roleID = rs.getInt(1);
        Role role = RoleImpl.getInstance(roleID);
        LOG.debug("getUserRoles() next role="+role);
        result.add(role);
      }      
      
      WorkspaceImpl workspace = WorkspaceImpl.getInstance(workspaceID);
      if (workspace instanceof UserWorkspace) {
        if (workspace==user.getUserWorkspace()) { // Note that we could compare references because of the cache!
          // and finally add the default Role.User
          Role role = RoleImpl.getInstance(Role.USER);
          LOG.debug("getUserRoles() next role="+role);
          result.add(role);      
        }
      }
    } catch (SQLException e) {
      LOG.error("getUserRoles("+user.getNick()+") FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }       
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }
    }  
    
    return result;
  }

  public Permission getPermission(String name) {
    return PermissionImpl.getInstance(name);
  }
  
  public Role getRole(String name) {
   return RoleImpl.getInstance(name); 
  }
  
  public List getUsersWithoutRoles(Workspace w) {
    LOG.debug("getUsersWithoutRoles()");
    List result = new ArrayList();
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT id, nick, password, name, lastname, email, info FROM "+USERS_TABLE+" WHERE id NOT IN " +
                                    "(SELECT DISTINCT user_id FROM "+ACCESS_RIGHTS_TABLE+" WHERE id = ?)");
      pstmt.setInt(1, w.getID());
      rs = pstmt.executeQuery();
      
      while (rs.next()) {
        int id = rs.getInt(1);
        String nick = rs.getString(2);
        String password = rs.getString(3);
        String name = rs.getString(4);
        String lastName = rs.getString(5);
        String email = rs.getString(6);
        String info = rs.getString(7);
        UserImpl user = UserImpl.getInstance(id, nick, password, name, lastName, email, info);
        result.add(user);
      }
      
    } catch (SQLException e) {
      LOG.error("getUsersWithoutRoles() FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return result;
    
  }

  public void deleteUserRole(User user, Role r, Workspace w) {
    String _method = "deleteUserRole("+user+","+r+","+w+") "; 
    LOG.debug(_method);
    Connection conn = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("DELETE FROM " + ACCESS_RIGHTS_TABLE + " WHERE id=? AND user_id=? AND role_id=?");
      int i = 0;
      pstmt.setInt(++i, w.getID());
      pstmt.setInt(++i, user.getID());
      pstmt.setInt(++i, r.getID());
      pstmt.executeUpdate();
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }      
  }

  public void deleteUserRoles(User user, List roles, Workspace w) {
    // DELETE FROM access_rights WHERE id=9 AND user_id=2 AND role_id IN (2,6)
    String _method = "deleteUserRoles("+user+","+w+","+roles+") "; 
    int rolesCount = roles.size();
    LOG.debug(_method);
    Connection conn = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("DELETE FROM " + ACCESS_RIGHTS_TABLE + " WHERE id=? AND user_id=? AND role_id IN ("+qmarks(rolesCount)+")");
      int i = 0;
      pstmt.setInt(++i, w.getID());
      pstmt.setInt(++i, user.getID());
      for (int j = 0; j < rolesCount; j++) {
        pstmt.setInt(++i, ((Role) roles.get(j)).getID());
      }
      pstmt.executeUpdate();
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }      
  }
  
  private static String qmarks(int count) {
    StringBuffer sb = new StringBuffer(2*count-1);
    for (int i = 1; i < count; i++) {
      sb.append('?').append(',');
    }
    sb.append('?');
    return sb.toString();
  }

  public boolean checkPermission(User user, int workspaceID, Permission p) {
    return _checkPermission(user, workspaceID, p);
  }

}
