package com.fmi.weprom.impl.mysql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import com.fmi.weprom.api.File;
import com.fmi.weprom.api.WEPROMException;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.*;

public class FileImpl implements File {
  
  private static final Log LOG = LogFactory.getLog(FileImpl.class);
  private static final String CACHE_NAME = "FileCache";
  private static final int NO_PARENT = -1;
  private static final String ROOT_FOLDER_NAME = "home"; 
  
  private String name;
  private byte type;
  private int id;
  private Integer _ID = null;
  
  int parent_id;
  int owner_id;
  int access_type;
  Date creation_date;
  long size;
  
  private transient List folders;
  private transient List files;
  
  
  private FileImpl(String name, byte type, int parent_id) {
    this(name, type, UNSAVED_ID, parent_id);
  }
  
  public FileImpl(String name, byte type, int id, int parent_id) {
    this.name = name;
    this.type = type;
    this.id = id;
    this.parent_id = parent_id;
  }
  
  public String getPath() {
    return null;
  }

  public String getName() {
    if (type==MASTER_TYPE_COMMUNITY || type==MASTER_TYPE_PROJECT || type==MASTER_TYPE_USERWORKSPACE) {
      return ROOT_FOLDER_NAME;
    }
    return name;
  }

  boolean isRootFolder() {
    return parent_id==NO_PARENT;
  }

  private static boolean isWorkspaceFolder(int type) {
    switch (type) {
      case MASTER_TYPE_USERWORKSPACE:
      case MASTER_TYPE_COMMUNITY:
      case MASTER_TYPE_PROJECT:
        return true;
        
      default:
        return false;
    }
  }
  
  public boolean isFolder() {
    switch (type) {
      case MASTER_TYPE_USERWORKSPACE:
      case MASTER_TYPE_COMMUNITY:
      case MASTER_TYPE_PROJECT:
      case MASTER_TYPE_FOLDER:
        return true;
        
      default:
        return false;
    }
  }

  public int getID() {
    return id;
  }

  static FileImpl getInstance(int id) {
    FileImpl file = getCached(id);
    if (file==null) {
      file = load(id);
    }
    return file;  
  }

  public List getFolders() {
    String _method = "getFolders() "; 
    LOG.debug(_method);
    if (folders==null) {
      folders = loadFolders(id); 
    }
    LOG.debug(_method+": " +folders);
    return folders;
  }

  public List getFiles() {
    String _method = "getFiles() "; 
    LOG.debug(_method);
    if (files==null) {
      files = loadFiles(id); 
    }
    LOG.debug(_method+": " +files);
    return files;
  }
  
  public int getParentID() {
    return parent_id;
  }

  private static FileImpl getCached(int id) {
    //LOG.debug("getCached id=" + id);
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    return (FileImpl) cache.get(new Integer(id));
  }

  private static List loadFiles(int parent_id) {
    return loadChildren(parent_id, File.TYPE_FILE, false, null, false, 0, 0);
  }
  
  private static List loadFolders(int parent_id) {
    return loadChildren(parent_id, File.TYPE_FOLDER, false, null, false, 0, 0);
  }
  
  public List getFiles(String sortColumn, boolean sortOrder, int offset, int pageSize) {
    String _method = "getFiles("+sortColumn+","+sortOrder+","+offset+","+pageSize+") "; 
    LOG.debug(_method);
    List list = loadChildren(id, File.TYPE_FILE, true, sortColumn, sortOrder, offset, pageSize); 
    LOG.debug(_method+": " +list);
    return list;
  }
  
  private static List loadChildren(int parent_id, int type, boolean extended, String sortColumn, boolean sortOrder, int offset, int pageSize) {
    List result = null;
    String _method = "loadChildren(" + parent_id + ") "; 
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    //int type = MASTER_TYPE_FOLDER;
    try {
      conn = Util.currentSession().getConnection();
      
      StringBuilder sql = new StringBuilder();
      sql.append("SELECT id,name,owner_id,access_type,creation_date,size ");
      sql.append("FROM ").append(MASTER_TABLE).append(" WHERE parent_id=? AND type=?");
      
      if (extended) {
        // select * from master where parent_id=17 and type=5 order by name desc limit 0,10;
        if (sortColumn!=null) {
          sql.append(" ORDER BY ?");
          if (sortOrder) {
            sql.append(" DESC");
          }
        }        
        sql.append(" LIMIT ?,?");
        pstmt = conn.prepareStatement(sql.toString());
        
        int i = 0;
        pstmt.setInt(++i, parent_id);
        pstmt.setInt(++i, type);
        if (sortColumn!=null) {
          pstmt.setString(++i, sortColumn);
        }
        pstmt.setInt(++i, offset);
        pstmt.setInt(++i, pageSize);
      } else {
        pstmt = conn.prepareStatement(sql.toString());
        int i = 0;
        pstmt.setInt(++i, parent_id);
        pstmt.setInt(++i, type);
      }

      LOG.debug(_method+"sql: "+pstmt.toString());
      rs = pstmt.executeQuery();
      result = new ArrayList();
      while (rs.next()) {
        int i = 0;
        int id = rs.getInt(++i);
        String name = rs.getString(++i);
        int owner_id = rs.getInt(++i);
        int access_type = rs.getInt(++i);
        Timestamp creation_dateTimestamp = rs.getTimestamp(++i);        
        long size = rs.getLong(++i);
        
        Date creation_date = null;
        if (creation_dateTimestamp!=null) {
          creation_date = new Date(creation_dateTimestamp.getTime());
        }
        
        FileImpl file = getInstance(
            id,
            parent_id,
            type,
            name,
            owner_id,
            access_type,
            creation_date,
            size);
        result.add(file);
        LOG.debug(_method+"next: "+file);
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return result;    
  }
  
  private static FileImpl load(int id) {
    String _method = "load(" + id + ") "; 
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT parent_id,type,name,owner_id,access_type,creation_date,size " +
                "FROM "+MASTER_TABLE+" WHERE id=?"); 
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      if (rs.next()) {
        int i = 0;
        int parent_id = rs.getInt(++i);
        boolean noParent = rs.wasNull();
        if (noParent) {
          parent_id = NO_PARENT;
        }
        
        int type = rs.getInt(++i);
        String name = rs.getString(++i);
        int owner_id = rs.getInt(++i);
        int access_type = rs.getInt(++i);
        Timestamp creation_dateTimestamp = rs.getTimestamp(++i);   
        long size = rs.getLong(++i);
        
        Date creation_date = null;
        if (creation_dateTimestamp!=null) {
          creation_date = new Date(creation_dateTimestamp.getTime());
        }
        
        FileImpl file = getInstance(
            id,
            parent_id,
            type,
            name,
            owner_id,
            access_type,
            creation_date,
            size);
        return file;
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;    
  }

  private static FileImpl getInstance(int id, int parent_id, int type, String name, int owner_id, int access_type, Date creation_date, long size) {
    FileImpl file = getCached(id);
    if (file==null) {
      file = new FileImpl(name, (byte) type, id, parent_id);
      file.owner_id = owner_id;
      file.access_type = access_type;
      file.creation_date = creation_date;
      file.size = size;
      file.cache();
    } else {
      file.name = name;
      file.parent_id = parent_id;
      file.owner_id = owner_id;
      file.access_type = access_type;
      file.creation_date = creation_date;
      file.size = size;
    }
    return file;
  }

  private void cache() {
    LOG.debug("cache : " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.put(get_ID(), this);
    
  }
  
  private Integer get_ID() {
    if (id==UNSAVED_ID) return null;
    if (_ID==null) {
      _ID = new Integer(id);
    }
    return _ID;
  }
  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("File[").append(id).append(',').append(name).append(']');
    return sb.toString();
  }

  public File createChild(byte type) {
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("File does not exists");
    }
    if (isFolder()) {
      return new FileImpl(null, type, id);
    }
    return null;
  }

  public void setName(String name) {
    this.name = name;
  }  
  
  public void save() throws WEPROMException {
    LOG.debug("save() ");
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("File already exists");
    }
    if (name==null || name.length()==0) {
      throw new IllegalStateException("File does not have a name");
    }
    
    boolean transactionOK = false;
    
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    int i = 0;
    try {
      conn = Util.currentSession().getConnection();
      
      LOG.debug("save() BEGIN TRANSACTION: CREATE_FILE");
      // Disable auto commit
      conn.setAutoCommit(false);
      
      if (isFileNameFree(conn, name, type, parent_id)) {
        // create file 
        pstmt = conn.prepareStatement("INSERT INTO "+MASTER_TABLE+" (parent_id,type,name,owner_id,access_type,creation_date,size)" +
            " VALUES(?,?,?,?,?,NOW(),?)", PreparedStatement.RETURN_GENERATED_KEYS);
        
        pstmt.setInt(++i, parent_id);
        pstmt.setInt(++i, type);
        pstmt.setString(++i, name);
        pstmt.setInt(++i, owner_id);
        pstmt.setInt(++i, access_type);
        pstmt.setLong(++i, size);
        
        pstmt.executeUpdate();
        
        // get autogenerated project id
        rs = pstmt.getGeneratedKeys();
        if (rs.next()) {
          int fileID = rs.getInt(1);
          this.id = fileID; 
          LOG.debug("save() generated id="+id);
          
          
          // Commit updates
          conn.commit();
          transactionOK = true;
          LOG.debug("save() END TRANSACTION: CREATE_FILE");       
        }
      } else {
        LOG.error("save() " + "FAILED: duplicate file name");
      }
    } catch (SQLException e) {
      LOG.error("save() " + "FAILED", e);      
      throw new WEPROMException("Faild to create file");
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("update() ROLLBACK TRANSACTION: CREATE_FILE");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}       
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }       
  }
  
  private boolean isFileNameFree(Connection conn, String fileName, int type, int parent_id) throws SQLException {
    PreparedStatement pstmt = conn.prepareStatement("SELECT * FROM "+MASTER_TABLE+" WHERE name=? AND type=? AND parent_id=?");
    int i = 0;
    pstmt.setString(++i, fileName);
    pstmt.setInt(++i, type);
    pstmt.setInt(++i, parent_id);
    ResultSet rs = pstmt.executeQuery();
    boolean result = !rs.next();
    LOG.debug("isFileNameFree("+fileName+"): "+result);
    return result;
  }

  private int filesCount = -1;
  private int foldersCount = -1;
  public int getFilesCount() throws WEPROMException {
    if (filesCount==-1) {
      filesCount = getChildrenCount(TYPE_FILE);
    }
    return filesCount;
  }
  
  public int getFoldersCount() throws WEPROMException {
    if (foldersCount==-1) {
      foldersCount = getChildrenCount(TYPE_FOLDER);
    } 
    return foldersCount;
  }
  
  private int getChildrenCount(int type) throws WEPROMException {
    String _method = "getChildrenCount() "+toString() + ' '; 
    LOG.debug(_method);
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    int count = 0;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT COUNT(*) FROM "+MASTER_TABLE+" WHERE parent_id=? AND type=?");
      int i = 0;
      pstmt.setInt(++i, id);
      pstmt.setInt(++i, type);      
      rs = pstmt.executeQuery();
      if (rs.next()) {
        count = rs.getInt(1);
      }
      LOG.debug(_method + ": " + count);
    } catch (SQLException e) {
      LOG.error(_method + "FAILED", e);      
      throw new WEPROMException("getUnreadMessageCount FAILED");
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return count;
    
  }

  public Date getDate() {
    return creation_date;
  }

  public void setDate(Date creation_date) {
    this.creation_date = creation_date;
  }
  
  public WorkspaceImpl getRootWorkspace() {
    FileImpl rootFile = getRoot();
    int _id = rootFile.getID();
    int _type = rootFile.type;
    
    switch (_type) {
      case MASTER_TYPE_USERWORKSPACE:
        UserImpl user = UserImpl.getInstance(rootFile.owner_id);
        return user.getUserWorkspace();
        
      case MASTER_TYPE_COMMUNITY:
        return CommunityImpl.getInstance(_id);
        
      case MASTER_TYPE_PROJECT:
        return ProjectImpl.getInstance(_id);
      
      default:
        return null;
    }
  }
  
  public FileImpl getRoot() {
    int _type = type;
    FileImpl _file = this;
    int _id = -1;
    while (!isWorkspaceFolder(_type)) {
      _id = _file.getParentID();
      _file = load(_id);
      _type = _file.type;
    }
    return _file;
  }

  public long getSize() {
    return size;
  }

  public void setSize(long size) {
    this.size = size;
  }
}
