package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.INVALID_ID;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MESSAGES_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.UNSAVED_ID;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.Formatter;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.Message;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.WEPROMException;

public class MessageImpl implements Message, Serializable {
  
  private static final Log LOG = LogFactory.getLog(MessageImpl.class);
  private static final String CACHE_NAME = "MessagesCache";
  static final String MESSAGES[] = {"TEXT", "PROJECT_INVITATION", "COMMUNITY_INVITATION", "TASK_ASSIGNMENT"};
  
  protected int id;
  
  protected transient User receiver;
  protected int receiver_id = INVALID_ID;
  
  protected transient User sender;
  protected int sender_id = INVALID_ID;
  
  protected String subject;
  protected String text;
  protected Date sentDate;
  protected int type = INVALID_ID;
  protected boolean isNew = true;
  boolean isReplied = false; 
  boolean isForwarded = false;
  boolean isInbox;
  
  ///////////////////////////////////////////////////////////////////////////////////////  
  boolean infoExtracted = false;
  int extractedProjectID = INVALID_ID;
  int extractedCommunityID = INVALID_ID;
  int extractedTaskID = INVALID_ID;
  static final String PROJECT_INVITATION_MSG_TEMPLATE = "User %s invites you to \r\n Project : \"%s\"\r\n";
  static final String COMMUNITY_INVITATION_MSG_TEMPLATE = "User %s invites you to \r\n Community : \"%s\"\r\n";
  static final String TASK_ASSIGNMENT_MSG_TEMPLATE = "User %s assigns to you:\r\n Task: \"%s\"\r\nin\r\n Project : \"%s\"\r\n";
  private static final String UNKNOWN = "UNKNOWN";
  private String assembledText = null;
  ///////////////////////////////////////////////////////////////////////////////////////
  
  
  MessageImpl(UserWorkspaceImpl userWorkspace) {
    this.id = UNSAVED_ID;
    this.sender = userWorkspace.getOwner();
    this.sender_id = sender.getID();
  }
  
  private MessageImpl(UserWorkspaceImpl userWorkspace, int id, int sender_id, int receiver_id, 
                      int type, String subject, String text, Date sentDate, boolean isInbox) {
    this.id = id;
    this.sender_id = sender_id;
    this.receiver_id = receiver_id;
    this.type = type;
    this.subject = subject;
    this.text = text;
    this.sentDate = sentDate;
    this.isInbox = isInbox;
  }

  public User getSender() {
    if (sender==null && sender_id!=INVALID_ID) {
      // lazy load the sender
      sender  = UserImpl.getInstance(sender_id);
    }
    return sender;    
  }
  
  public User getReceiver() {
    if (receiver==null && receiver_id!=INVALID_ID) {
      // lazy load the receiver
      receiver  = UserImpl.getInstance(receiver_id);
    }
    return receiver;    
  }

  public int getID() {
    return id;
  }

  public void accept() throws WEPROMException {
    LOG.debug("accept()");
    switch (type) {
      case Message.MESSAGE_TYPE_PROJECT_INVITATION:
        MessageProcessor.acceptProjectInvitation(this);
        break;
        
      case Message.MESSAGE_TYPE_COMMUNITY_INVITATION:
        MessageProcessor.acceptCommunityInvitation(this);
        break;
        
      case Message.MESSAGE_TYPE_TASK_ASSIGNMENT:
        MessageProcessor.acceptTaskAssignment(this);
        break;
        
      default:
        throw new IllegalStateException("Accept is not applicable to this type of message : " + type);
    }
  }

  public void decline() throws WEPROMException {
    LOG.debug("decline()");
    switch (type) {
    case Message.MESSAGE_TYPE_PROJECT_INVITATION:
      MessageProcessor.declineProjectInvitation(this);
      break;
      
    case Message.MESSAGE_TYPE_COMMUNITY_INVITATION:
      MessageProcessor.declineCommunityInvitation(this);
      break;
      
    case Message.MESSAGE_TYPE_TASK_ASSIGNMENT:
      MessageProcessor.declineTaskAssignment(this);
      break;
      
    default:
      throw new IllegalStateException("Decline is not applicable to this type of message : " + type);
  }    
  }

  public void send() throws WEPROMException {
    save();
  }

  public void delete() throws WEPROMException {
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Message does not exists");
    }    
    LOG.debug("delete()");
    Connection conn = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("DELETE FROM "+ MESSAGES_TABLE+" WHERE id=?");
      pstmt.setInt(1, id);
      pstmt.executeUpdate();      
    } catch (SQLException e) {
      throw new WEPROMException("delete() FAILED", e);
    } finally {
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    } 
    uncache();
  }  
  
  static List loadInboxMessages(UserWorkspaceImpl uw) throws WEPROMException {
    return loadMessages(uw, true);
  }

  static List loadOutboxMessages(UserWorkspaceImpl uw) throws WEPROMException {
    return loadMessages(uw, false);
  }
  
  private static List loadMessages(UserWorkspaceImpl uw, boolean inbox) throws WEPROMException {
    String method = null;
    if (inbox) {
      method = "loadMessages(inbox) ";
    } else {
      method = "loadMessages(outbox) ";
    }
    LOG.debug(method);
    List result = new ArrayList();
    
    int sender_id = -1;
    int receiver_id = -1;
    if (inbox) {
      receiver_id = uw.getOwner().getID();
    } else {
      sender_id = uw.getOwner().getID();
    }
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    
    try {
      conn = Util.currentSession().getConnection();
      String sql = null;
      if (inbox) {
        sql = "SELECT id,sender_id,type,subject,text,date,new,replied,forwarded FROM " + MESSAGES_TABLE +
              " WHERE receiver_id=? AND inbox=?";
      } else {
        sql = "SELECT id,receiver_id,type,subject,text,date,new,replied,forwarded FROM " + MESSAGES_TABLE +
        " WHERE sender_id=? AND inbox=?";        
      }
      pstmt = conn.prepareStatement(sql);
      pstmt.setInt(1, ( inbox ? receiver_id : sender_id));
      pstmt.setBoolean(2, inbox);
      rs = pstmt.executeQuery();
      
      while (rs.next()) {
        int i = 0;
        int id = rs.getInt(++i);
        if (inbox) {
          sender_id = rs.getInt(++i);
        } else {
          receiver_id = rs.getInt(++i);
        }
        int type = rs.getInt(++i);
        String subject = rs.getString(++i);
        String text = rs.getString(++i);
        Timestamp t = rs.getTimestamp(++i);
        Date sentDate = new Date(t.getTime());
        boolean isNew = rs.getBoolean(++i);
        boolean isReplied = rs.getBoolean(++i);
        boolean isForwarded = rs.getBoolean(++i);
        
        MessageImpl msg = getInstance(uw, id, sender_id, receiver_id, type, subject, text, sentDate, isNew, isReplied, isForwarded, inbox);
        LOG.debug(method+"next msg="+msg);
        result.add(msg);
      }
    } catch (SQLException e) {
      throw new WEPROMException(method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }    
    return result;
  }
  
  private static MessageImpl load(UserWorkspaceImpl uw, int id) {
    String _method = "load(" + id + ") ";   
    LOG.debug(_method);
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    
    try {
      conn = Util.currentSession().getConnection();
      String sql = null;     
      sql = "SELECT sender_id,receiver_id,type,subject,text,date,new,replied,forwarded,inbox FROM " + MESSAGES_TABLE +
            " WHERE id=?";
      
      pstmt = conn.prepareStatement(sql);
      pstmt.setInt(1, id);
      rs = pstmt.executeQuery();
      
      if (rs.next()) {
        int i = 0;        
        int sender_id = rs.getInt(++i);
        int receiver_id = rs.getInt(++i);
        int type = rs.getInt(++i);
        String subject = rs.getString(++i);
        String text = rs.getString(++i);
        Timestamp t = rs.getTimestamp(++i);
        Date sentDate = new Date(t.getTime());
        boolean isNew = rs.getBoolean(++i);
        boolean isReplied = rs.getBoolean(++i);
        boolean isForwarded = rs.getBoolean(++i);
        boolean inbox = rs.getBoolean(++i);
        
        MessageImpl msg = getInstance(uw, id, sender_id, receiver_id, type, subject, text, sentDate, isNew, isReplied, isForwarded, inbox);
        LOG.debug(_method+": "+msg);
        return msg;
      }
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);;
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }    
    return null;
  }
  
  static int getUnreadMessageCount(UserWorkspaceImpl uw) throws WEPROMException {
    LOG.debug("getUnreadMessageCount() ");
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    int count = 0;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("SELECT COUNT(*) FROM "+MESSAGES_TABLE+" WHERE receiver_id=? AND inbox=? AND new=?");
      int i = 0;
      pstmt.setInt(++i, uw.getOwner().getID());
      pstmt.setBoolean(++i, true);
      pstmt.setBoolean(++i, true);
      rs = pstmt.executeQuery();
      if (rs.next()) {
        count = rs.getInt(1);
      }
      LOG.debug("getUnreadMessageCount() " + ": " + count);
    } catch (SQLException e) {
      LOG.error("getUnreadMessageCount() " + "FAILED", e);      
      throw new WEPROMException("getUnreadMessageCount FAILED");
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return count;
  }
  
  // note that this method participates in other transactions, so it should not commit!
  public void update() throws WEPROMException {
    LOG.debug("update() ");
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("Could not update unsaved Message");
    }
    
    Connection conn = null;
    ResultSet rs = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("UPDATE "+MESSAGES_TABLE+" SET new=?,replied=?,forwarded=? WHERE id=? AND inbox=?");
      int i = 0;
      pstmt.setBoolean(++i, isNew);
      pstmt.setBoolean(++i, isReplied);
      pstmt.setBoolean(++i, isForwarded);
      pstmt.setInt(++i, id);
      pstmt.setBoolean(++i, isInbox);
      pstmt.executeUpdate();
    } catch (SQLException e) {
      LOG.error("update() " + "FAILED", e);      
      throw new WEPROMException("Faild to update task");
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }           
  }
  
  void save() throws WEPROMException {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Message already sent");
    }
    if (subject==null) {
      throw new IllegalStateException("Message does not have a subject");
    }
    if (receiver_id==INVALID_ID) {
      throw new IllegalStateException("Message does not have a receiver");
    }
    if (type==INVALID_ID) {
      throw new IllegalStateException("Message does not have a type");
    }
    LOG.debug("save()");
    
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    
    try {
      conn = Util.currentSession().getConnection();
      String sql = "INSERT INTO "+MESSAGES_TABLE + " (receiver_id,sender_id,type,subject,text,date,new,replied,forwarded,inbox)" +
          " VALUES(?,?,?,?,?,NOW(),?,?,?,?)";
      pstmt = conn.prepareStatement(sql, PreparedStatement.RETURN_GENERATED_KEYS);
      int i = 0;
      pstmt.setInt(++i, receiver_id);
      pstmt.setInt(++i, sender_id);
      pstmt.setInt(++i, type);
      pstmt.setString(++i, subject);
      pstmt.setString(++i, text);
      pstmt.setBoolean(++i, isNew);
      pstmt.setBoolean(++i, isReplied);
      pstmt.setBoolean(++i, isForwarded);
      pstmt.setBoolean(++i, false);
      pstmt.executeUpdate();
      
      // get autogenerated project id
      rs = pstmt.getGeneratedKeys();
      if (rs.next()) {
        int msgID = rs.getInt(1);
        this.id = msgID; 
        LOG.debug("save() generated id="+id);
      } else {
        throw new WEPROMException("Failed to obtain the generated id when saving Message : " + toString());
      }
      
      // make the copy for the reciever
      pstmt.setBoolean(i, true);
      pstmt.executeUpdate();
      
    } catch (SQLException e) {
      throw new WEPROMException("Failed to save Message : " + toString(), e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }
    }
  }
  
  static MessageImpl getInstance(
      UserWorkspaceImpl uw,
      int id,
      int sender_id,
      int receiver_id,
      int type,
      String subject,
      String text,
      Date sentDate,
      boolean isNew, 
      boolean isReplied, 
      boolean isForwarded,
      boolean isInbox) {
    
    MessageImpl msg = getCached(id);
    if (msg==null) {
      msg = new MessageImpl(uw, id, sender_id, receiver_id, type, subject, text, sentDate, isInbox);
      LOG.debug("getInstance(...) create : " + msg);
      msg.cache();
    } else {
      // messages are immutable
      LOG.debug("getInstance(...) loaded from cache : " + msg);
    }
    msg.setNew(isNew);
    msg.setReplied(isReplied);
    msg.setForwarded(isForwarded);
    
    return msg;
  }
  
  static MessageImpl getInstance(UserWorkspaceImpl uw, int id) {
    MessageImpl message = getCached(id);
    if (message==null) {
      message = load(uw, id);
    }
    return message;
  }
  
  private static MessageImpl getCached(int id) {
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    return (MessageImpl) cache.get(new Integer(id));
  }  
  
  private void uncache() {
    LOG.debug("uncache: " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.remove(get_ID());
  }
  
  private void cache() {
    LOG.debug("cache : " + toString());
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    cache.put(get_ID(), this);
  }
  
  private Integer _ID = null;
  private Integer get_ID() {
    if (id==UNSAVED_ID) return null;
    if (_ID==null) {
      _ID = new Integer(id);
    }
    return _ID;
  }  

  public String getSubject() {
    return subject;
  }

  public void setSubject(String subject) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Could not update "+"subject"+" of existing message");
    }    
    this.subject = subject;
  }

  public String getBody() {
    switch (type) {
      case MESSAGE_TYPE_PROJECT_INVITATION:
        return assembleProjectInvitationMsgText();
      
      case MESSAGE_TYPE_COMMUNITY_INVITATION:
        return assembleCommunityInvitationMsgText();

      case MESSAGE_TYPE_TASK_ASSIGNMENT:
        return assembleTaskAsignementMsgText();

      default:
        return text;
    }
  }

  public void setBody(String text) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Could not update "+"text"+" of existing message");
    }    
    this.text = text;
  }

  public int getType() {
    return type;
  }

  public void setType(int type) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Could not update "+"type"+" of existing message");
    }    
    this.type = type;
  }

  public Date getDate() {
    return sentDate;
  }

  public void setReceiver(User receiver) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Could not update "+"receiver"+" of existing message");
    }      
    this.receiver_id = receiver.getID();
    this.receiver = receiver;
  }

  public void setSender(User sender) {
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("Could not update "+"sender"+" of existing message");
    }      
    this.sender_id = sender.getID();
    this.sender = sender;
  }
  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Message[id=").append(id);
    
    sb.append(',');
    if (type>=0 && type<MESSAGES.length) {
      sb.append(MESSAGES[type]);
    } else {
      sb.append("unknow type");
    }
    
    if (sender!=null) {
      sb.append(",sender="+sender.getNick());
    } else {
      sb.append(",sender_id="+sender_id);
    }
    if (receiver!=null) {
      sb.append(",receiver="+receiver.getNick());
    } else {
      sb.append(",receiver_id="+receiver_id);
    }    
    sb.append(',').append('\"').append(subject).append('\"');
    sb.append(']');
    return sb.toString();
  }
  
  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();    
    LOG.debug("SERIALIZE: " + toString());    
  }
  
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();    
    LOG.debug("DESERIALIZE: " + toString());    
  }

  public void setNew(boolean b) {
    isNew = b;
  }
  
  public boolean isNew() {
    return isNew;
  }

  public boolean isForwarded() {
    return isForwarded;
  }

  public void setForwarded(boolean b) {
    isForwarded = b;
  }

  public boolean isReplied() {
    return isReplied;
  }

  public void setReplied(boolean b) {
    isReplied = b;
  }

  public boolean isInbox() {
    return isInbox;
  }
  
  int getProjectID() {
    return -1;
  }
  
  int getTaskID() {
    return -1;
  }
  
  int getReceiverID() {
    return receiver_id;
  }  

  ///////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////
  
  void extractInfo() {
    if (infoExtracted) return;
    String _method = "extractInfo() ";
    
    switch (type) {
      case MESSAGE_TYPE_PROJECT_INVITATION:
        try {
          extractedProjectID = Integer.parseInt(text);
        } catch (NumberFormatException e) {
          LOG.debug(_method+"FAILED to parse project id ("+text+")");
        }        
        break;
        
      case MESSAGE_TYPE_COMMUNITY_INVITATION:
        try {
          extractedCommunityID = Integer.parseInt(text);
        } catch (NumberFormatException e) {
          LOG.debug(_method+"FAILED to parse community id ("+text+")");
        }           
        break;
        
      case MESSAGE_TYPE_TASK_ASSIGNMENT:
        Scanner scanner = new Scanner(text);
        scanner.useDelimiter(",");
        if (scanner.hasNextInt()) {
          extractedProjectID = scanner.nextInt();
          if (scanner.hasNextInt()) {
            extractedTaskID = scanner.nextInt();
          } else {
            LOG.debug(_method+"FAILED to extract task id from "+toString());  
          }
        } else {
          LOG.debug(_method+"FAILED to extract project id from "+toString());
        }
        break;
        
      default:
        break;
    }
    
    infoExtracted = true;
  }
  
  int extractProjectID() {
    extractInfo();
    return extractedProjectID;
  }
  
  int extractCommunityID() {
    extractInfo();
    return extractedCommunityID;
  }  

  int extractTaskID() {
    extractInfo();
    return extractedTaskID;
  }
  
  static MessageImpl createCommunityInvitationMessage(CommunityImpl community, User host, User guest) {
    LOG.debug("createCommunityInvitationMessage("+community+", "+host+" : "+guest+")");
    MessageImpl msg = new MessageImpl((UserWorkspaceImpl) host.getUserWorkspace());
    String communityName = community.getName();
    int communityID = community.getID();
    msg.setSubject("Community Invitation : "+communityName);
    msg.setBody(String.valueOf(communityID));
    msg.setType(Message.MESSAGE_TYPE_COMMUNITY_INVITATION);
    msg.setReceiver(guest);
    return msg;
  }
  
  static MessageImpl createProjectInvitationMessage(ProjectImpl project, User host, User guest) {
    LOG.debug("createProjectInvitationMessage("+project+", "+host+" : "+guest+")");
    MessageImpl msg = new MessageImpl((UserWorkspaceImpl) host.getUserWorkspace());
    String projectName = project.getName();
    int projectID = project.getID();
    msg.setSubject("Project Invitation : "+projectName);
    msg.setBody(String.valueOf(projectID));
    msg.setType(Message.MESSAGE_TYPE_PROJECT_INVITATION);
    msg.setReceiver(guest);
    return msg;
  }
  
  private String assembleCommunityInvitationMsgText() {
    if (assembledText==null) {
      StringBuilder sb = new StringBuilder();
      Formatter formatter = new Formatter(sb); // new Formatter(sb, Locale.US);
      
      String communityName = "UNKNOWN";     
      int communityID = extractCommunityID();      
      if (communityID!=INVALID_ID) {
        CommunityImpl community = CommunityImpl.getInstance(communityID);
        if (community!=null) {
          communityName = community.getName();
        }
      }
      
      String invitorName = "UNKNOWN";
      if (sender_id != INVALID_ID) {
        UserImpl invitor = UserImpl.getInstance(sender_id);
        if (invitor!=null) {
          invitorName = invitor.getNick();
        }
      }
      
      formatter.format(COMMUNITY_INVITATION_MSG_TEMPLATE, invitorName, communityName);
      formatter.flush();
      assembledText = sb.toString();
    }
    return assembledText;
  }
  
  private String assembleProjectInvitationMsgText() {
    if (assembledText==null) {
      StringBuilder sb = new StringBuilder();
      Formatter formatter = new Formatter(sb); // new Formatter(sb, Locale.US);
      
      String projectName = "UNKNOWN";     
      int projectID = extractProjectID();      
      if (projectID!=INVALID_ID) {
        ProjectImpl project = ProjectImpl.getInstance(projectID);
        if (project!=null) {
          projectName = project.getName();
        }
      }
      
      String invitorName = "UNKNOWN";
      if (sender_id != INVALID_ID) {
        UserImpl invitor = UserImpl.getInstance(sender_id);
        if (invitor!=null) {
          invitorName = invitor.getNick();
        }
      }
      
      formatter.format(PROJECT_INVITATION_MSG_TEMPLATE, invitorName, projectName);
      formatter.flush();
      assembledText = sb.toString();
    }
    return assembledText;
  }
  
  static MessageImpl createTaskAsignementMessage(UserWorkspaceImpl userWorkspace, TaskImpl task) {
    MessageImpl msg = new MessageImpl(userWorkspace);
    String taskSubject = task.getSubject();
    msg.setSubject("Task Assignement: "+taskSubject);
    int taskID = task.getID();
    ProjectImpl project = task.getProject();
    int projectID = project.getID();
    //String projectName = project.getName();
    
    StringBuilder sb = new StringBuilder();
    sb.append(projectID);
    sb.append(',').append(taskID);
    msg.setBody(sb.toString());
    msg.setType(Message.MESSAGE_TYPE_TASK_ASSIGNMENT);
    msg.setReceiver(task.getProcessor());
    return msg;
  }
  
  private String assembleTaskAsignementMsgText() {
    if (assembledText==null) {
      StringBuilder sb = new StringBuilder();
      Formatter formatter = new Formatter(sb); // new Formatter(sb, Locale.US);
      
      String projectName = UNKNOWN;     
      int projectID = extractProjectID();      
      if (projectID!=INVALID_ID) {
        ProjectImpl project = ProjectImpl.getInstance(projectID);
        if (project!=null) {
          projectName = project.getName();
        }
      }
      
      String invitorName = UNKNOWN;
      if (sender_id != INVALID_ID) {
        UserImpl invitor = UserImpl.getInstance(sender_id);
        if (invitor!=null) {
          invitorName = invitor.getNick();
        }
      }
      
      String taskSubject = UNKNOWN;
      int taskID = extractTaskID();
      if (taskID!=INVALID_ID) {
        TaskImpl task = TaskImpl.getInstance(taskID);
        if (task!=null) {
          taskSubject = task.getSubject();
        }
      }
      
      formatter.format(TASK_ASSIGNMENT_MSG_TEMPLATE, invitorName, taskSubject, projectName);
      formatter.flush();
      assembledText = sb.toString();
    }
    return assembledText;
  }
    
}
