package com.fmi.weprom.impl.mysql;

import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.ACCESS_RIGHTS_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.INVALID_ID;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TABLE;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_COMMUNITY;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.MASTER_TYPE_PROJECT;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.UNSAVED_ID;
import static com.fmi.weprom.impl.mysql.WEPROMSystemImpl.USERS_TABLE;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fmi.weprom.api.AccessRights;
import com.fmi.weprom.api.User;
import com.fmi.weprom.api.WEPROMException;
import com.fmi.weprom.api.Workspace;

public class UserImpl extends User {
  
  private static final Log LOG = LogFactory.getLog(UserImpl.class);
  
  private static final String CACHE_NAME = "UserCache";
  private static final String USER_ID_CACHE = "UserIDCache";
  
  private Integer _ID = null;
  
  public UserImpl(WEPROMSystemImpl weprom, String nick) {
    this(UNSAVED_ID, nick);
  }
  
  private UserImpl(int id, String nick) {
    this.id = id;
    this.nick = nick;
  }
  
  private UserImpl(int id, String nick, String password, String name, 
                    String lastName, String email, String information) {
    this.id = id;    
    this.nick = nick;
    this.password = password;
    this.name = name;
    this.lastName = lastName;
    this.email = email;
    this.info = information;
  }  
  
  public UserWorkspaceImpl getUserWorkspace() {
    return UserWorkspaceImpl.getInstance(this);
  }

  public void save() throws WEPROMException {
    LOG.debug("save() id="+id);
    if (id!=UNSAVED_ID) {
      throw new IllegalStateException("User already exists");
    }
    Connection conn = null;
    //Statement stmt = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    boolean transactionOK = false;
    try {
      conn = Util.currentSession().getConnection();
      String sql = "INSERT INTO users (nick,password,name,lastname,email,info,phone) VALUES(?,?,?,?,?,?,?)";
      pstmt = conn.prepareStatement(sql, PreparedStatement.RETURN_GENERATED_KEYS);
      int i = 0;
      pstmt.setString(++i, nick);
      pstmt.setString(++i, password);
      pstmt.setString(++i, name);
      pstmt.setString(++i, lastName);
      pstmt.setString(++i, email);
      pstmt.setString(++i, info);
      pstmt.setString(++i, phone);
      
      LOG.debug("save() BEGIN TRANSACTION: CREATE_USER_AND_WORKSPACE");
      // Disable auto commit
      conn.setAutoCommit(false);
      
      pstmt.executeUpdate();
      
      // get autogenerated user id
      rs = pstmt.getGeneratedKeys();
      if (rs.next()) {
        int userID = rs.getInt(1);
        this.id = userID; 
        LOG.debug("save() generated id="+id);
        LOG.debug("save() save the corresponding UserWorkspace ...");
        UserWorkspaceImpl uw = new UserWorkspaceImpl(this); 
        if (uw!=null) {
          uw.save();  
          
          // Commit updates
          conn.commit();
          transactionOK = true;
          LOG.debug("save() END TRANSACTION: CREATE_USER_AND_WORKSPACE");
          cache();
        } else {
         LOG.error("save() user "+getNick() + " has no UserWorkspace."); 
        }         
      } else {
        LOG.debug("save() FAILED for "+toString());
      }      
    } catch (SQLException e) {
      //LOG.error("save() FAILED for "+toString(), e);
      throw new WEPROMException("save() FAILED for "+toString() ,e);
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("save() ROLLBACK TRANSACTION: CREATE_USER_AND_WORKSPACE");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }      
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }    
  }

  public void update() {
    LOG.debug("update() id="+id);
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("User does not exists");
    }    
    Connection conn = null;
    PreparedStatement pstmt = null;
    try {
      conn = Util.currentSession().getConnection();
      pstmt = conn.prepareStatement("UPDATE "+WEPROMSystemImpl.USERS_TABLE+
          " SET password=?,name=?,lastname=?,email=?,info=?,phone=?" +
          " WHERE id=?");
      
      int i = 0;
      pstmt.setString(++i, password);
      pstmt.setString(++i, name);
      pstmt.setString(++i, lastName);
      pstmt.setString(++i, email);
      pstmt.setString(++i, info);
      pstmt.setString(++i, phone);
      pstmt.setInt(++i, id);
      
      int updateCount = pstmt.executeUpdate();
      
    } catch (SQLException e) {
      LOG.error("update() FAILED", e);
    } finally {
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }    
  }

  public void delete() {
    LOG.debug("delete() id="+id);
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("User does not exists");
    }      
    boolean transactionOK = false;
    Connection conn = null;
    Statement stmt = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement();
      StringBuffer sb = new StringBuffer();
      sb.append("DELETE FROM").append(' ').append(WEPROMSystemImpl.USERS_TABLE).append(' ');
      sb.append("WHERE").append(' ').append("id").append('=').append('\'').append(id).append('\'');
      String sql = sb.toString();
      LOG.debug("delete() "+sql);
      
      LOG.debug("save() BEGIN TRANSACTION: DELETE_USER_AND_WORKSPACE");
      // Disable auto commit
      conn.setAutoCommit(false);
      
      int updateCount = stmt.executeUpdate(sql);
      
      UserWorkspaceImpl uw = getUserWorkspace();
      uw.delete();
      
      // Commit updates
      conn.commit();
      transactionOK = true;
      LOG.debug("save() END TRANSACTION: DELETE_USER_AND_WORKSPACE");      
    } catch (SQLException e) {
      LOG.error("delete()", e);
    } finally {
      if (!transactionOK) {
        try {
          LOG.debug("save() ROLLBACK TRANSACTION: DELETE_USER_AND_WORKSPACE");
          conn.rollback();
        } catch (SQLException ignore) {}
      }
      try {
        conn.setAutoCommit(true);
      } catch (SQLException ignore) {}      
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    uncache();
  }

  public List getCommunities() {
    LOG.debug("getCommunities() "+"for user "+toString());
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("User does not exist");
    }    
    return CommunityImpl.getUserCommunities(this);
  }

  public List getProjects() {
    LOG.debug("getProjects() "+"for user "+nick);
    if (id==UNSAVED_ID) {
      throw new IllegalStateException("User does not exist");
    }    
    return ProjectImpl.getUserProjects(this);
    /*
    List result = null;
    Connection conn = null;
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      //select distinct ar.id from access_rights as ar, master as m where user_id=3 and type=2
      //pstmt = conn.prepareStatement("SELECT DISTINCT ar.id FROM " + ACCESS_RIGHTS_TABLE + " AS ar, "+MASTER_TABLE+" AS m WHERE ar.user_id=? AND m.type=?");      
      pstmt = conn.prepareStatement("SELECT DISTINCT ar.id FROM " + ACCESS_RIGHTS_TABLE + " AS ar INNER JOIN "+MASTER_TABLE+" ma ON ma.id=ar.id WHERE ar.user_id=? AND ma.type=?");
      
      pstmt.setInt(1, id);
      pstmt.setInt(2, MASTER_TYPE_PROJECT);
      rs = pstmt.executeQuery();
      
      result = new ArrayList();
      while (rs.next()) {
        int workspace_id = rs.getInt(1);
        LOG.debug("getProjects() "+"next workspace_id="+workspace_id);
        ProjectImpl p = ProjectImpl.getInstance(workspace_id);
        result.add(p);
      }
    } catch (SQLException e) {
      LOG.error("getProjects() "+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }   
      if (pstmt!=null) {
        try {
          pstmt.close();
        } catch (SQLException ignore) {}
      }      
    }      
    return result;
    */
  }

  static UserImpl getInstance(int id) {
    UserImpl user = getCached(id);
    if (user!=null) {
      return user;
    }
    String _method = "getInstance("+id+") ";//, weprom="+weprom+") "; 
    LOG.debug(_method);
    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement(); 
      rs = stmt.executeQuery("SELECT nick, password, name, lastname, email, info FROM "+USERS_TABLE+" WHERE id='"+id+"'");
      if (rs.next()) {
        String nick = rs.getString(1);
        String password = rs.getString(2);
        String name = rs.getString(3);
        String lastName = rs.getString(4);
        String email = rs.getString(5);
        String info = rs.getString(6);
        
        user = UserImpl.getInstance(id, nick, password, name, lastName, email, info);
        return user;
      }      
    } catch (SQLException e) {
      LOG.error(_method+"FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;
  }
  
  static User getInstance(String nick) {
    UserImpl user = getCached(nick);
    if (user!=null) {
      return user;
    }    
    LOG.debug("getInstance("+nick+")");
    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement(); 
      rs = stmt.executeQuery("SELECT id, password, name, lastname, email, info FROM "+USERS_TABLE+" WHERE nick='"+nick+"'");
      if (rs.next()) {
        int id = rs.getInt(1);
        String password = rs.getString(2);
        String name = rs.getString(3);
        String lastName = rs.getString(4);
        String email = rs.getString(5);
        String info = rs.getString(6);
        user = getInstance(id, nick, password, name, lastName, email, info);
        return user;
      }      
    } catch (SQLException e) {
      LOG.error("getInstance() FAILED to get user with nick="+nick, e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return null;
  }  
  
  static List getUsers() {
    LOG.debug("getUsers()");
    List result = new ArrayList();
    Connection conn = null;
    Statement stmt = null;
    ResultSet rs = null;
    try {
      conn = Util.currentSession().getConnection();
      stmt = conn.createStatement(); 
      rs = stmt.executeQuery("SELECT id, nick, password, name, lastname, email, info FROM "+USERS_TABLE);
      while (rs.next()) {
        int id = rs.getInt(1);
        String nick = rs.getString(2);
        String password = rs.getString(3);
        String name = rs.getString(4);
        String lastName = rs.getString(5);
        String email = rs.getString(6);
        String info = rs.getString(7);
        User user = getInstance(id, nick, password, name, lastName, email, info);
        result.add(user);
      }
      
    } catch (SQLException e) {
      LOG.error("getUsers() FAILED", e);
    } finally {
      if (rs!=null) {
        try {
          rs.close();
        } catch (SQLException ignore) {}
      }
      if (stmt!=null) {
        try {
          stmt.close();
        } catch (SQLException ignore) {}
      }      
    }
    return result;
  }  
  
  public AccessRights getAccessRights(Workspace w) {
    // TODO Auto-generated method stub
    return null;
  }

  private void uncache() {
    LOG.debug("uncache: " + toString());
    Map userCache = Util.currentSession().getCache(CACHE_NAME);
    userCache.remove(nick);

    userCache = Util.currentSession().getCache(USER_ID_CACHE);
    userCache.remove(get_ID());
  }
  
  private void cache() {
    LOG.debug("cache: " + toString());
    Map userCache = Util.currentSession().getCache(CACHE_NAME);
    userCache.put(nick, this);
    
    userCache = Util.currentSession().getCache(USER_ID_CACHE);
    userCache.put(get_ID(), this);
  }
  
  private static UserImpl getCached(int id) {
    Map cache = Util.currentSession().getCache(USER_ID_CACHE);
    return (UserImpl) cache.get(new Integer(id));
  }
  
  private static UserImpl getCached(String nick) {
    Map cache = Util.currentSession().getCache(CACHE_NAME);
    return (UserImpl) cache.get(nick);
  }
  
  
  static UserImpl getInstance(int id, String nick, String password, String name, String lastName, String email, String info) {
    UserImpl user = getCached(nick);    
    if (user==null) {
      LOG.debug("create new User object: " + nick);
      user = new UserImpl(id, nick, password, name, lastName, email, info);
      user.cache();//userCache.put(nick, user);
    } else {
      LOG.debug("update cached User object: " + nick);
      user.setPassword(password);
      user.setName(name);
      user.setLastName(lastName);
      user.setEmail(email);
    }
    return user;    
  }

  private Integer get_ID() {
    if (_ID==null && id!=INVALID_ID) {
      _ID = new Integer(id);
    }
    return _ID;
  }
  
  private void writeObject(ObjectOutputStream oos) throws IOException {
    oos.defaultWriteObject();    
    LOG.debug("SERIALIZE: " + toString());    
  }
  
  private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
    ois.defaultReadObject();    
    LOG.debug("DESERIALIZE: " + toString());    
  }    
}
